==> spread-5.7.0.72-py3-none-any.whl: spread_core/bam/dali/providers/providers.py <==
import logging
from typing import Generator

import spread_core.errors.project_errors as errors
import spread_dali as dali
from .. import bindings
from ...entities import Funit
from ...providers import Provider


class DaliProvider(Provider):
    _dali_type = 0
    dali_type = property(fget=lambda self: self._dali_type)

    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliProvider, self).__init__(id, **kwargs)
        self._binding = binding
        self._status = self.SUSPICIOUS

    @staticmethod
    def parse_args(attributes: dict, **kwargs) -> dict:
        return dict(
            binding=bindings.of(attributes)
        )

    @property
    def rank(self):
        return self.dali_type, 0, self.id

    def __str__(self):
        return super(DaliProvider, self).__str__() + f' {self._binding}'

    def groups(self):
        """
        Список групповых провайдеров, в группы которых включён provider
        """
        for item in self.get_value(Groups, []):
            if item['value']:
                yield from self.manager.group_providers(self.dali_type, item['index'])

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == Discovery:
            return False

        return super(DaliProvider, self).is_repaired_funit(funit_type)

    @property
    def surveyable(self):
        if self.is_device:
            return super(DaliProvider, self).surveyable

        return False

    @property
    def binding(self) -> dali.address.Address:
        return self._binding

    @property
    def short_addr(self) -> dali.address.Short:
        assert isinstance(self._binding, dali.address.Short)
        return self._binding

    @property
    def is_device(self):
        return isinstance(self.binding, dali.address.Short)

    @property
    def is_group(self):
        return isinstance(self.binding, dali.address.Group)

    @property
    def is_broadcast(self):
        return isinstance(self.binding, dali.address.Broadcast)

    @property
    def is_unaddressed(self):
        return isinstance(self.binding, dali.address.BroadcastUnaddressed)

    def get_members(self, return_self=False):
        if self.is_group:
            yield from self.manager.get_group_members(self.dali_type, self._binding.group)
            if return_self:
                yield self
        elif self.is_broadcast:
            yield from self.manager.get_all_devices_of_type(self.dali_type)
            if return_self:
                yield self
        else:
            yield self

    def dtr(self, index):
        return self._manager.get_dtr_value(self._dali_type, index)

    def request(self, cmd: dali.command.Command, *args, **kwargs):
        try:
            sig = kwargs.pop('sig', cmd.sig)
            response = self._manager.send_interface.send(cmd, *args, **kwargs)
            if cmd.is_query:
                cmd_args = []

                if isinstance(cmd, dali.gear.general._GearCommand) and cmd._hasparam:
                    cmd_args.append(cmd.param)

                if cmd._uses_dtr0:
                    cmd_args.append(self.dtr(0))

                if cmd._uses_dtr1:
                    cmd_args.append(self.dtr(1))

                if cmd._uses_dtr2:
                    cmd_args.append(self.dtr(2))

                [self.on_request_answer(funit_type, response, *cmd_args, sig=sig) for funit_type in cmd.funit_type]
        except BaseException as ex:
            logging.exception(ex)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if isinstance(response, dali.command.YesNoResponse):
            self.on_update(funit_type, response.value, sig=sig)
        elif funit_type == 'MemoryPart' and response:
            self.on_update(*_MemoryValue.update_part(self, response.value.as_integer, *args), sig=sig)
        elif isinstance(response, dali.command.Response):
            funit = DaliFunit.of(funit_type)
            if funit and issubclass(funit, DaliFunit):
                if funit.index_of:
                    value = funit.index_of[response.value.as_integer]
                elif funit.mask:
                    value = [funit.mask[i] for i in range(0, 8) if response.value.as_integer & 2**i != 0]
                else:
                    value = response.value.as_integer

                if funit.value_type:
                    if value.__class__ != funit.value_type:
                        if funit.value_type == list:
                            value = [value]
                        else:
                            value = funit.value_type(value)
            else:
                value = response.value.as_integer

            self.on_update(funit_type, value, sig=sig)
        else:
            Provider.on_update(self, funit_type, None, sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == Validity:
            if self.is_valid and not value:
                self.set_valid(self.SUSPICIOUS)
            elif self._status != self.VALID and value:
                self.set_valid(self.VALID)
        elif funit_type == '%s0-7' % Groups:
            self._parse_groups(0, value, sig=sig)
        elif funit_type == '%s8-15' % Groups:
            self._parse_groups(8, value, sig=sig)
        elif funit_type == '%s16-23' % Groups:
            self._parse_groups(16, value, sig=sig)
        elif funit_type == '%s24-31' % Groups:
            self._parse_groups(24, value, sig=sig)
        else:
            if funit_type == Groups and value is not None and self.is_device:
                for item in value:
                    if item['value']:
                        self.manager.add_to_group(self, item['index'])
                    else:
                        self.manager.remove_from_group(self, item['index'])

            super().on_update(funit_type, value, sig, **kwargs)

    def get_info(self):
        if Discovery not in self.values:
            self.on_update(Discovery, False)
        if Groups not in self.values:
            self.GetGroups()

    def on_survey(self):
        yield from super(DaliProvider, self).on_survey()

    def on_external(self, command: dali.command.Command, response, *args):
        if command.is_query:
            for funit_type in command.funit_type:
                self.on_request_answer(funit_type, response, *args)

        elif self.is_valid:
            self.on_command_sended(command)

    def update_recursive(self, funit_type, value, sig=None, **kwargs):
        self.on_update(funit_type, value, sig, **kwargs)

        if not self.is_device:
            for _provider in self.get_members():
                _provider.on_update(funit_type, value, sig=None, **kwargs)

    def check_command_for_device(self):
        if not self.is_device:
            raise errors.JocketError('The GET command cannot be applied to an object that is not an DEVICE!')

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        raise NotImplementedError()

    def SetDiscovery(self, value, sig=None):
        raise NotImplementedError()

    def _parse_groups(self, first_index: int, value, sig=None):
        assert first_index % 8 == 0, "Incorrect first index"

        _groups = [{key: value for key, value in item.items()} for item in self.get_value(Groups, [])]
        if len(_groups) != Groups.length[self._dali_type]:
            _groups = Groups.def_value(self._dali_type)

        for _i in range(0, 8):
            _groups[first_index + _i]['value'] = bool(value & (1 << _i))

        self.update_recursive(Groups, _groups, sig=sig)

    def GetGroups(self, _=None, sig=None, **kwargs):
        _groups = [{'index': self._binding.group, 'value': True}] if self.is_group else []
        _groups = self.get_value(Groups) or _groups

        self.on_update(Groups, _groups, sig=sig)

    def SetGroups(self, val, sig=None):
        raise NotImplementedError()

    def serialize(self) -> dict:
        d = {}
        if self.is_device:
            d['address'] = self.binding.address
            for funit_type in [Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion]:
                value = self.get_value(funit_type)
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                if value is not None:
                    d[key] = value
        return d


class DaliFunit(Funit):
    get_for_not_devices = True
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None
    easy_get = False
    get_after_set = True
    _dtr0 = None
    _dtr1 = None
    _dtr2 = None

    @property
    def entity(self) -> DaliProvider:
        return self._entity

    def _prepare_cmd(self, provider, cmd, *args):
        raise NotImplemented()

    def _apply_cmd(self, provider: DaliProvider, cmd, *args):
        i = 0

        if cmd._uses_dtr0:
            yield self._dtr0(args[i])
            i += 1

        if cmd._uses_dtr1:
            yield self._dtr1(args[i])
            i += 1

        if cmd._uses_dtr2:
            yield self._dtr2(args[i])
            i += 1

        yield self._prepare_cmd(provider, cmd, *args[i:])

    def set(self):
        if self.set_cmd:
            for cmd in self._apply_cmd(self.entity, self.set_cmd, self.set_value, *self.params):
                self.entity.request(cmd, sig=self.signature)

            if not self.entity.is_device:
                self.entity.on_update(self.funit_type, self._variable.value)

        elif self.set_func:
            getattr(self.entity, self.set_func.__name__)(self.set_value, **self.params, sig=self.signature)

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        else:
            if not self.entity.is_device:
                # return
                if self.get_for_not_devices:
                    self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
                    self._variable.clear_signature()
                else:
                    return

            if self.get_cmd or self.get_func:
                for _provider in self.entity.get_members():
                    if isinstance(_provider, self.entity_type):
                        if self.get_cmd:
                            for cmd in self._apply_cmd(_provider, self.get_cmd):
                                _provider.request(cmd, sig=self.signature)
                        else:
                            assert self.get_func
                            getattr(_provider, self.get_func.__name__)(**self.params, sig=self.signature)


DaliProvider.funit_class = DaliFunit


class Discovery(DaliFunit):
    UP_CMD = dali.gear.general.RecallMaxLevel
    DOWN_CMD = dali.gear.general.Off
    get_for_not_devices = False
    easy_get = True
    set_func = DaliProvider.SetDiscovery


class Groups(DaliFunit):
    length = {1: 16, 2: 32}
    @staticmethod
    def def_value(dali_type) -> list:
        return {
            1: [{'index': i, 'value': False} for i in range(0, Groups.length[1])],
            2: [{'index': i, 'value': False} for i in range(0, Groups.length[2])]
        }.get(dali_type)

    set_func = DaliProvider.SetGroups
    get_func = DaliProvider.GetGroups


class _MemoryValue(DaliFunit):
    get_for_not_devices = False
    bank_id = 0
    start = None
    end = None

    def set(self):
        super(DaliFunit, self).set()

    def get(self):
        if self.entity.is_device:
            self.entity.on_update(
                self.funit_type,
                self.to_string(self.entity.read_memory_location(self.bank_id, self.start, self.end)),
                sig=self.signature
            )

        else:
            super(DaliFunit, self).get()

    @classmethod
    def as_hexed(cls, _bytes: bytes) -> str:
        return bytes(_bytes).hex().upper()

    @classmethod
    def as_int(cls, _bytes: bytes) -> str:
        return '.'.join([str(b) for b in _bytes])

    to_string = as_hexed

    @classmethod
    def update_part(cls, entity: DaliProvider, value: int, location_id: int, bank_id: int) -> (DaliFunit, str):
        for _f in (Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion):
            if bank_id == _f.bank_id and _f.start <= location_id < _f.end:
                index = location_id-_f.start
                if _f.to_string == _f.as_hexed:
                    v = entity.get_value(_f, '..'*(_f.end - _f.start))
                    v = v[:index*2] + bytes([value]).hex().upper() + v[(index+1)*2:(_f.end - _f.start)*2]
                    return _f, v
                else:
                    v = entity.get_value(_f, '.').split('.')
                    v[index] = str(value)
                    return _f, '.'.join(v)

        raise ValueError('Unknown memory part of [%s, %s]' % (bank_id, location_id))


class Gtin(_MemoryValue):
    start = 0x3
    end = 0x9


class Serial(_MemoryValue):
    start = 0xb
    end = 0x13


class GtinOem(_MemoryValue):
    bank_id = 1
    start = 0x3
    end = 0x9


class SerialOem(_MemoryValue):
    bank_id = 1
    start = 0x9
    end = 0x11


class FirmwareVersion(_MemoryValue):
    start = 0x9
    end = 0xb
    to_string = _MemoryValue.as_int


class HardwareVersion(_MemoryValue):
    start = 0x13
    end = 0x15
    to_string = _MemoryValue.as_int


class Validity(DaliFunit):
    easy_get = True

==> spread-5.9.0.59-py3-none-any.whl: spread_core/bam/dali/providers/providers.py <==
import logging
from typing import Generator

import spread_core.errors.project_errors as errors
import spread_dali as dali
from .. import bindings
from ...entities import Funit
from ...providers import Provider


class DaliProvider(Provider):
    _dali_type = 0
    dali_type = property(fget=lambda self: self._dali_type)
    _dali_device_type = 0
    dali_device_type = property(fget=lambda self: self._dali_device_type)

    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliProvider, self).__init__(id, **kwargs)
        self._binding = binding
        self._status = self.SUSPICIOUS

    @staticmethod
    def parse_args(attributes: dict, **kwargs) -> dict:
        return dict(
            binding=bindings.of(attributes),
            name=kwargs.pop('name', '')
        )

    @property
    def rank(self):
        return self.dali_type, 0, self.id

    def __str__(self):
        return super(DaliProvider, self).__str__() + f' {self._binding}'

    def groups(self):
        """
        Список групповых провайдеров, в группы которых включён provider
        """
        for item in self.get_value(Groups, []):
            if item['value']:
                yield from self.manager.group_providers(self.dali_type, item['index'])

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == Discovery:
            return False

        return super(DaliProvider, self).is_repaired_funit(funit_type)

    @property
    def surveyable(self):
        if self.is_device:
            return super(DaliProvider, self).surveyable

        return False

    @property
    def binding(self) -> dali.address.Address:
        return self._binding

    @property
    def short_addr(self) -> dali.address.Short:
        assert isinstance(self._binding, dali.address.Short)
        return self._binding

    @property
    def is_device(self):
        return isinstance(self.binding, dali.address.Short)

    @property
    def is_group(self):
        return isinstance(self.binding, dali.address.Group)

    @property
    def is_broadcast(self):
        return isinstance(self.binding, dali.address.Broadcast)

    @property
    def is_unaddressed(self):
        return isinstance(self.binding, dali.address.BroadcastUnaddressed)

    def get_members(self, return_self=False):
        if self.is_group:
            yield from self.manager.get_group_members(self.dali_type, self._binding.group)
            if return_self:
                yield self
        elif self.is_broadcast:
            yield from self.manager.get_all_devices_of_type(self.dali_type)
            if return_self:
                yield self
        else:
            yield self

    def dtr(self, index):
        return self._manager.get_dtr_value(self._dali_type, index)

    def request(self, cmd: dali.command.Command, *args, **kwargs):
        try:
            sig = kwargs.pop('sig', cmd.sig)
            response = self._manager.send_interface.send(cmd, *args, **kwargs)
            if cmd.is_query:
                cmd_args = []

                if isinstance(cmd, dali.gear.general._GearCommand) and cmd._hasparam:
                    cmd_args.append(cmd.param)

                if cmd._uses_dtr0:
                    cmd_args.append(self.dtr(0))

                if cmd._uses_dtr1:
                    cmd_args.append(self.dtr(1))

                if cmd._uses_dtr2:
                    cmd_args.append(self.dtr(2))

                [self.on_request_answer(funit_type, response, *cmd_args, sig=sig) for funit_type in cmd.funit_type]
        except BaseException as ex:
            logging.exception(ex)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if isinstance(response, dali.command.YesNoResponse):
            self.on_update(funit_type, response.value, sig=sig)
        elif funit_type == 'MemoryPart' and response:
            self.on_update(*_MemoryValue.update_part(self, response.value.as_integer, *args), sig=sig)
        elif isinstance(response, dali.command.Response):
            funit = DaliFunit.of(funit_type)
            if funit and issubclass(funit, DaliFunit):
                if funit.index_of:
                    value = funit.index_of[response.value.as_integer]
                elif funit.mask:
                    value = [funit.mask[i] for i in range(0, 8) if response.value.as_integer & 2**i != 0]
                else:
                    value = response.value.as_integer

                if funit.value_type:
                    if value.__class__ != funit.value_type:
                        if funit.value_type == list:
                            value = [value]
                        else:
                            value = funit.value_type(value)
            else:
                value = response.value.as_integer

            self.on_update(funit_type, value, sig=sig)
        else:
            Provider.on_update(self, funit_type, None, sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == Validity:
            if self.is_valid and not value:
                self.set_valid(self.SUSPICIOUS)
            elif self._status != self.VALID and value:
                self.set_valid(self.VALID)
        elif funit_type == '%s0-7' % Groups:
            self._parse_groups(0, value, sig=sig)
        elif funit_type == '%s8-15' % Groups:
            self._parse_groups(8, value, sig=sig)
        elif funit_type == '%s16-23' % Groups:
            self._parse_groups(16, value, sig=sig)
        elif funit_type == '%s24-31' % Groups:
            self._parse_groups(24, value, sig=sig)
        else:
            if funit_type == Groups and value is not None and self.is_device:
                for item in value:
                    if item['value']:
                        self.manager.add_to_group(self, item['index'])
                    else:
                        self.manager.remove_from_group(self, item['index'])

            super().on_update(funit_type, value, sig, **kwargs)

    def get_info(self):
        if Discovery not in self.values:
            self.on_update(Discovery, False)
        if Groups not in self.values:
            self.GetGroups()

    def on_survey(self):
        yield from super(DaliProvider, self).on_survey()

    def on_external(self, command: dali.command.Command, response, *args):
        if command.is_query:
            for funit_type in command.funit_type:
                self.on_request_answer(funit_type, response, *args)

        elif self.is_valid:
            self.on_command_sended(command)

    def update_recursive(self, funit_type, value, sig=None, **kwargs):
        self.on_update(funit_type, value, sig, **kwargs)

        if not self.is_device:
            for _provider in self.get_members():
                _provider.on_update(funit_type, value, sig=None, **kwargs)

    def check_command_for_device(self):
        if not self.is_device:
            raise errors.JocketError('The GET command cannot be applied to an object that is not an DEVICE!')

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        raise NotImplementedError()

    def SetDiscovery(self, value, sig=None):
        raise NotImplementedError()

    def _parse_groups(self, first_index: int, value, sig=None):
        assert first_index % 8 == 0, "Incorrect first index"

        _groups = [{key: value for key, value in item.items()} for item in self.get_value(Groups, [])]
        if len(_groups) != Groups.length[self._dali_type]:
            _groups = Groups.def_value(self._dali_type)

        for _i in range(0, 8):
            _groups[first_index + _i]['value'] = bool(value & (1 << _i))

        self.update_recursive(Groups, _groups, sig=sig)

    def GetGroups(self, _=None, sig=None, **kwargs):
        _groups = [{'index': self._binding.group, 'value': True}] if self.is_group else []
        _groups = self.get_value(Groups) or _groups

        self.on_update(Groups, _groups, sig=sig)

    def SetGroups(self, val, sig=None):
        raise NotImplementedError()

    def serialize(self) -> dict:
        d = {}
        if self.is_device:
            d['address'] = self.binding.address
            for funit_type in [Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion]:
                value = self.get_value(funit_type)
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                if value is not None:
                    d[key] = value
        return d


class DaliFunit(Funit):
    get_for_not_devices = True
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None
    easy_get = False
    get_after_set = True
    _dtr0 = None
    _dtr1 = None
    _dtr2 = None

    @property
    def entity(self) -> DaliProvider:
        return self._entity

    def _prepare_cmd(self, provider, cmd, *args):
        raise NotImplemented()

    def _apply_cmd(self, provider: DaliProvider, cmd, *args):
        i = 0

        if cmd._uses_dtr0:
            yield self._dtr0(args[i])
            i += 1

        if cmd._uses_dtr1:
            yield self._dtr1(args[i])
            i += 1

        if cmd._uses_dtr2:
            yield self._dtr2(args[i])
            i += 1

        yield self._prepare_cmd(provider, cmd, *args[i:])

    def set(self):
        if self.set_cmd:
            for cmd in self._apply_cmd(self.entity, self.set_cmd, self.set_value, *self.params):
                self.entity.request(cmd, sig=self.signature)

            if not self.entity.is_device:
                self.entity.on_update(self.funit_type, self._variable.value)

        elif self.set_func:
            getattr(self.entity, self.set_func.__name__)(self.set_value, **self.params, sig=self.signature)

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        else:
            if not self.entity.is_device:
                # return
                if self.get_for_not_devices:
                    self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
                    self._variable.clear_signature()
                else:
                    return

            if self.get_cmd or self.get_func:
                for _provider in self.entity.get_members():
                    if isinstance(_provider, self.entity_type):
                        if self.get_cmd:
                            for cmd in self._apply_cmd(_provider, self.get_cmd):
                                _provider.request(cmd, sig=self.signature)
                        else:
                            assert self.get_func
                            getattr(_provider, self.get_func.__name__)(**self.params, sig=self.signature)


DaliProvider.funit_class = DaliFunit


class Discovery(DaliFunit):
    UP_CMD = dali.gear.general.RecallMaxLevel
    DOWN_CMD = dali.gear.general.Off
    get_for_not_devices = False
    easy_get = True
    set_func = DaliProvider.SetDiscovery


class Groups(DaliFunit):
    length = {1: 16, 2: 32}
    @staticmethod
    def def_value(dali_type) -> list:
        return {
            1: [{'index': i, 'value': False} for i in range(0, Groups.length[1])],
            2: [{'index': i, 'value': False} for i in range(0, Groups.length[2])]
        }.get(dali_type)

    set_func = DaliProvider.SetGroups
    get_func = DaliProvider.GetGroups


class _MemoryValue(DaliFunit):
    get_for_not_devices = False
    bank_id = 0
    start = None
    end = None

    def set(self):
        super(DaliFunit, self).set()

    def get(self):
        if self.entity.is_device:
            self.entity.on_update(
                self.funit_type,
                self.to_string(self.entity.read_memory_location(self.bank_id, self.start, self.end)),
                sig=self.signature
            )

        else:
            super(DaliFunit, self).get()

    @classmethod
    def as_hexed(cls, _bytes: bytes) -> str:
        return bytes(_bytes).hex().upper()

    @classmethod
    def as_int(cls, _bytes: bytes) -> str:
        return '.'.join([str(b) for b in _bytes])

    to_string = as_hexed

    @classmethod
    def update_part(cls, entity: DaliProvider, value: int, location_id: int, bank_id: int) -> (DaliFunit, str):
        for _f in (Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion):
            if bank_id == _f.bank_id and _f.start <= location_id < _f.end:
                index = location_id-_f.start
                if _f.to_string == _f.as_hexed:
                    v = entity.get_value(_f, '..'*(_f.end - _f.start - 1))
                    v = v[:index*2] + bytes([value]).hex().upper() + v[(index+1)*2:(_f.end - _f.start)*2]
                    return _f, v
                else:
                    v = entity.get_value(_f, '.').split('.')
                    v[index] = str(value)
                    return _f, '.'.join(v)

        raise ValueError('Unknown memory part of [%s, %s]' % (bank_id, location_id))


class Gtin(_MemoryValue):
    start = 0x3
    end = 0x9


class Serial(_MemoryValue):
    start = 0xb
    end = 0x13


class GtinOem(_MemoryValue):
    bank_id = 1
    start = 0x3
    end = 0x9


class SerialOem(_MemoryValue):
    bank_id = 1
    start = 0x9
    end = 0x11


class FirmwareVersion(_MemoryValue):
    start = 0x9
    end = 0xb
    to_string = _MemoryValue.as_int


class HardwareVersion(_MemoryValue):
    start = 0x13
    end = 0x15
    to_string = _MemoryValue.as_int


class Validity(DaliFunit):
    easy_get = True
